# -*- coding: utf-8 -*-
"""
JSON encoder/decoder adapted for use with Google App Engine NDB.

Usage:

  import ndb_json
  
  # Serialize an ndb.Query into an array of JSON objects.
  query = models.MyModel.query()
  query_json = ndb_json.dumps(query)
  
  # Convert into a list of Python dictionaries.
  query_dicts = ndb_json.loads(query_json)
  
  # Serialize an ndb.Model instance into a JSON object.
  entity = query.get()
  entity_json = ndb_json.dumps(entity)
  
  # Convert into a Python dictionary.
  entity_dict = ndb_json.loads(entity_json)


Dependencies:

  - dateutil: https://pypi.python.org/pypi/python-dateutil
"""
 
__author__ = 'Eric Higgins'
__copyright__ = 'Copyright 2013, Eric Higgins'
__version__ = '0.0.5'
__email__ = 'erichiggins@gmail.com'


import base64
import datetime
import json
import re
import time
import types

import dateutil.parser
from google.appengine.ext import ndb


__all__ = (
    'dump',
    'dumps',
    'loads',
)


def encode_model(obj):
  """Encode objects like ndb.Model which have a `.to_dict()` method."""
  obj_dict = obj.to_dict()
  for key, val in obj_dict.iteritems():
    if isinstance(val, types.StringType):
      try:
        unicode(val)
      except UnicodeDecodeError:
        # Encode binary strings (blobs) to base64.
        obj_dict[key] = base64.b64encode(val)
  return obj_dict


def encode_generator(obj):
  """Encode generator-like objects, such as ndb.Query."""
  return list(obj)


def encode_key(obj):
  """Get the Entity from the ndb.Key for further encoding."""
  # NOTE(erichiggins): Potentially poor performance for Models w/ many KeyProperty properties.
  return obj.get_async()
  # Alternative 1: Convert into pairs.
  # return obj.pairs()
  # Alternative 2: Convert into URL-safe base64-encoded string.
  # return obj.urlsafe()


def encode_future(obj):
  """Encode an ndb.Future instance."""
  return obj.get_result()


def encode_datetime(obj):
  """Encode a datetime.datetime or datetime.date object as an ISO 8601 format string."""
  # Reformat the date slightly for better JS compatibility.
  # Offset-naive dates need 'Z' appended for JS.
  # datetime.date objects don't have or need tzinfo, so don't append 'Z'.
  zone = '' if getattr(obj, 'tzinfo', True) else 'Z'
  return obj.isoformat() + zone


def encode_complex(obj):
  """Convert a complex number object into a list containing the real and imaginary values."""
  return [obj.real, obj.imag]


def encode_basevalue(obj):
  """Retrieve the actual value from a ndb.model._BaseValue.
  
  This is a convenience function to assist with the following issue:
  https://code.google.com/p/appengine-ndb-experiment/issues/detail?id=208
  """
  return obj.b_val


NDB_TYPE_ENCODING = {
  ndb.MetaModel: encode_model,
  ndb.Query: encode_generator,
  ndb.QueryIterator: encode_generator,
  ndb.Key: encode_key,
  ndb.Future: encode_future,
  datetime.date: encode_datetime,
  datetime.datetime: encode_datetime,
  time.struct_time: encode_generator,
  types.ComplexType: encode_complex,
  ndb.model._BaseValue: encode_basevalue,
  
}


class NdbEncoder(json.JSONEncoder):
  """Extend the JSON encoder to add support for NDB Models."""

  def default(self, obj):
    """Overriding the default JSONEncoder.default for NDB support."""

    obj_type = type(obj)
    # NDB Models return a repr to calls from type().
    if obj_type not in NDB_TYPE_ENCODING and hasattr(obj, '__metaclass__'):
      obj_type = obj.__metaclass__
    fn = NDB_TYPE_ENCODING.get(obj_type)
    if fn:
      return fn(obj)

    return json.JSONEncoder.default(self, obj)


def dumps(ndb_model, **kwargs):
  """Custom json dumps using the custom encoder above."""
  return NdbEncoder(**kwargs).encode(ndb_model)


def dump(ndb_model, fp, **kwargs):
  """Custom json dump using the custom encoder above."""
  for chunk in NdbEncoder(**kwargs).iterencode(ndb_model):
    fp.write(chunk)


def loads(json_str, **kwargs):
  """Custom json loads function that converts datetime strings."""
  json_dict = json.loads(json_str, **kwargs)
  if isinstance(json_dict, list):
    return map(iteritems, json_dict)
  return iteritems(json_dict)


def iteritems(json_dict):
  """Loop over a json dict and try to convert strings to datetime."""
  for key, val in json_dict.iteritems():
    if isinstance(val, dict):
      iteritems(val)
    # TODO(erichiggins): Find a better way to detect date/time-like strings.
    # Its a little hacky to check for specific chars, but avoids integers.
    elif isinstance(val, basestring) and val.count('-') == 2 and len(val) > 9:
      try:
        json_dict[key] = dateutil.parser.parse(val)
        # Check for UTC.
        if val.endswith(('+00:00', '-00:00', 'Z')):
          # Then remove tzinfo for gae, which is offset-naive.
          json_dict[key] = json_dict[key].replace(tzinfo=None)
      except (TypeError, ValueError):
        pass
  return json_dict
  



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_ndb_json
----------------------------------

Tests for `gaek` module.
"""

import datetime
import json
import unittest
import cStringIO

from gaek import ndb_json


class TestNdbJson(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_dumps_with_ndb_values(self):
        """Assert functional with NDB datastore values."""
        pass

    def test_dumps_with_naive_values(self):
        """Assert functional parity with `json.dump` for naive values."""
        assert json.dumps({'bool true': True}) == ndb_json.dumps({'bool true': True})
        assert json.dumps({'bool false': False}) == ndb_json.dumps({'bool false': False})
        assert json.dumps({'empty str': ''}) == ndb_json.dumps({'empty str': ''})
        assert json.dumps({'str': 'hello world'}) == ndb_json.dumps({'str': 'hello world'})
        assert json.dumps({'unicode': u'(╯°□°)╯︵ ┻━┻'}) == ndb_json.dumps({'unicode': u'(╯°□°)╯︵ ┻━┻'})
        assert json.dumps({'int': 12345}) == ndb_json.dumps({'int': 12345})
        assert json.dumps({'float': 1.2345}) == ndb_json.dumps({'float': 1.2345})
        assert json.dumps({'str int': '12345'}) == ndb_json.dumps({'str int': '12345'})
        assert json.dumps({'str float': '1.2345'}) == ndb_json.dumps({'str float': '1.2345'})

    def test_dump_with_naive_values(self):
        """Assert functional parity with `json.dump`, for use with file-like objects."""
        json_fp = cStringIO.StringIO()
        ndb_json_fp = cStringIO.StringIO()

        payload = {
            'unicode': u'(╯°□°)╯︵ ┻━┻',
        }

        json.dump(payload, json_fp)
        ndb_json.dump(payload, ndb_json_fp)

        assert json_fp.getvalue() == ndb_json_fp.getvalue()
        json_fp.close()
        ndb_json_fp.close()

    def test_loads_with_naive_values(self):
        payload_str = json.dumps({
            'bool true': True,
            'bool false': False,
            'empty str': '',
            'str': 'hello world',
            'unicode': u'(╯°□°)╯︵ ┻━┻',
            'int': 12345,
            'float': 1.2345,
            'str int': '12345',
            'str float': '1.2345',
        })

        parsed = ndb_json.loads(payload_str)

        assert True == parsed['bool true']
        assert False == parsed['bool false']
        assert '' == parsed['empty str']
        assert 'hello world' == parsed['str']
        assert u'(╯°□°)╯︵ ┻━┻' == parsed['unicode']
        assert 12345 == parsed['int']
        assert 1.2345 == parsed['float']
        assert '12345' == parsed['str int']
        assert '1.2345' == parsed['str float']

    def test_loads_with_date_and_time_values(self):
        """Assert that date/time-like strings are parsed properly."""
        payload_str = json.dumps({
            'date': datetime.date(2015, 1, 1).isoformat(),
            'datetime': datetime.datetime(2015, 1, 1, 12, 0, 0, 0).isoformat(),
            'time': '12:30',
            'non-date': '12-15',
            'double-hyphen non-date': '12-15-0',
        })

        parsed = ndb_json.loads(payload_str)

        # TODO(erichiggins): It may be better for dates to be parsed as instances of `datetime.date`.
        assert datetime.datetime(2015, 1, 1) == parsed['date']
        assert datetime.datetime(2015, 1, 1, 12, 0, 0, 0) == parsed['datetime']
        assert '12:30' == parsed['time']
        assert '12-15' == parsed['non-date']
        assert '12-15-0' == parsed['double-hyphen non-date']


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from setuptools import setup
from setuptools import find_packages


with open('VERSION') as version_file:
    VERSION = version_file.readline().rstrip('\n')

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read()

with open('test_requirements.txt') as test_requirements_file:
    test_requirements = test_requirements_file.read()

setup(
    name='gaek',
    version=VERSION,
    description="A collection of useful tools for Python apps running on Google App Engine.",
    author="Eric Higgins",
    author_email='erichiggins@gmail.com',
    url='https://github.com/erichiggins/gaek',
    packages=find_packages(exclude=['tests']),
    install_requires=requirements,
    license="BSD",
    keywords='gaek',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
